/*
    dsproxy_audio.c - OSS trampoline device driver

    Copyright (C) 2000 Ram Sivrasubramanium <ram@linux-workshop.com>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#define __NO_VERSION__

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/delay.h>
#include <linux/errno.h>
#include <linux/version.h>
#include <linux/string.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <linux/fs.h>
#include <linux/poll.h>
#include <asm/poll.h>
#include "dsproxy.h"


/* ########## VVVV                                                  VVVV  ########## */
/* ########## VVVV  needs to be called from em->sem protected code  VVVV  ########## */
/* ########## VVVV                                                  VVVV  ########## */

/*
 * fifo management/query functions
 */
static int get_fifo_avail(
	struct dsproxy_s *em)
{
	if (em->r_inuse) {
		return(em->bufsiz - em->fifocount);
	} else {
		return(em->bufsiz);  /* fake it to complete any pending writes */
	}
}

static void fifo_emit_char(
	struct dsproxy_s *em, 
	char ch)
{
	if (em->r_inuse) {
		em->vmbuf[em->fifoput] = ch;
		em->fifoput++;
		if (em->fifoput == DSPROXY_FIFO_SIZE) em->fifoput = 0;
		em->fifocount++;
	}
}


static void fifo_emit_int(
	struct dsproxy_s *em, 
	int val)
{
	int i;
	
	for (i = 0; i < sizeof(int); i++) {
		fifo_emit_char(em, (char)(val & 0xff));
		val >>= 8;
	}
}

static void fifo_emit_short(
	struct dsproxy_s *em, 
	int val)
{
	int i;
	
	for (i = 0; i < sizeof(short); i++) {
		fifo_emit_char(em, (char)(val&0xff));
		val >>= 8;
	}
}

/* ########## ^^^^                                                  ^^^^  ########## */
/* ########## ^^^^  needs to be called from em->sem protected code  ^^^^  ########## */
/* ########## ^^^^                                                  ^^^^  ########## */



/*
 * low water fifo v. freq management (requires to be called from em->ioctl_sem protected code!)
 */
static void dsproxy_calc_low_water(
	struct dsproxy_s *em)
{
	int val = em->freq / 10;
	
	if (val > DSPROXY_FRAG_SIZ) {
		val = DSPROXY_FRAG_SIZ;
	} else if (val < 256) {
		val = 256;
	}
	em->lowwater = val;
}


/*
 * calc 1/2 second fragment (requires to be called from em->ioctl_sem protected code!)
 */
static int dsproxy_calc_oss_optfrag(
	struct dsproxy_s *em)
{
	int i;
	
	if (! em->frag_explicitly_set) {
		em->fragsize = (em->channels * em->bytes_per_sample * em->freq) >> 1;  /* one half-second */
		if (! em->fragsize) em->fragsize = 1;
		
		em->numfrags = DSPROXY_FRAGFIFO_SIZE / em->fragsize;
		if (em->numfrags) {
			em->bufsiz = em->numfrags * em->fragsize;
		} else {
			em->bufsiz = em->fragsize = DSPROXY_FRAGFIFO_SIZE;
			em->numfrags = 1;
		}
	}
	
	for (i = 0; i < 14; i++) {
		if ((1 << i) > em->fragsize) break;
	}
	i--; 
	if (i < 0) i = 0;
	return((em->numfrags << 16) | i);
}


/*
 * dsp ioctl()
 */ 
static int dsproxy_audio_emit_ioctl_2(
	int typ, 
	struct dsproxy_s *em, 
	unsigned int cmd, 
	unsigned int val)
{
	int rc = 0;
	const int hdrsiz = (sizeof(char)) + (sizeof(int)) + (sizeof(int));  /* header + cmd + val */
	
	down(&em->sem);
	
	if (get_fifo_avail(em) - hdrsiz >= 0) {
		rc = 1;
		em->hdrskip += (const unsigned long long)hdrsiz;
		fifo_emit_char(em, typ);
		fifo_emit_int (em, cmd);
		fifo_emit_int (em, val);
	}
	
	up(&em->sem);
	return(rc);
}


int dsproxy_audio_emit_ioctl(
	int typ, 
	struct file *file, 
	struct dsproxy_s *em, 
	unsigned int cmd, 
	unsigned int val)
{
	DEBUG(printk("dsproxy_audio.o: dsproxy_audio_emit_ioctl \n"));
	for(;;) {
		if (dsproxy_audio_emit_ioctl_2(typ, em, cmd, val)) {
			wake_up_interruptible(&em->inq);
			break;
		} else {
			if (file->f_flags&O_NONBLOCK) return(-EAGAIN);
			interruptible_sleep_on(&em->outq);
			if (signal_pending(current)) return(-ERESTARTSYS);
		}
	} 
	
	return(0);
}


int dsproxy_audio_ioctl(
	struct file *file, 
	struct dsproxy_s *em, 
	unsigned int cmd, 
	unsigned long arg)
{
	int err, len = 0;
	int val;
	struct count_info c_info;
	struct timeval now;
	
	if (_SIOC_DIR(cmd) != _SIOC_NONE && _SIOC_DIR(cmd) != 0) {
		/*
		 * Have to validate the address given by the process.
		 */
		len = _SIOC_SIZE(cmd);
		if (len < 1 || len > 65536 || arg == 0) return(-EFAULT);
		if ((_SIOC_DIR(cmd) & _SIOC_WRITE) && ((err = verify_area(VERIFY_READ, (void *)arg, len)) < 0)) return(err);
		if ((_SIOC_DIR(cmd) & _SIOC_READ) && ((err = verify_area(VERIFY_WRITE, (void *)arg, len)) < 0)) return(err);
	}
	
	DEBUG(printk("ioctl: %08x\n", cmd));
	
	if (em->send_reset) {  /* implicitly send reset on device open */
		down(&em->ioctl_sem);
		em->send_reset = 0;
		up(&em->ioctl_sem);
		
		if (cmd != SNDCTL_DSP_RESET) {
			dsproxy_audio_emit_ioctl(DSPROXY_STREAM_IOCTL, file, em, SNDCTL_DSP_RESET, 0);
		}
	}
	
	switch (cmd) {
		case SNDCTL_DSP_RESET:
		{
			dsproxy_audio_flush(em, file);
			
			down(&em->ioctl_sem);
			em->format = AFMT_U8;
        	em->freq = 8000;
        	em->numfrags = 2;
        	em->fragsize = DSPROXY_FRAGFIFO_SIZE  / em->numfrags;
        	em->frag_explicitly_set = 0;
        	dsproxy_calc_low_water(em);
        	em->bytes_processed = 0ULL;
        	em->channels = 1;
			em->bytes_per_sample = 1;
			em->hdrskip = 0ULL;
			em->firstwrite.tv_sec = 0;
			em->firstwrite.tv_usec = 0;
			up(&em->ioctl_sem);
			
			dsproxy_audio_emit_ioctl(DSPROXY_STREAM_IOCTL, file, em, cmd, 0);
			return(0);
		}
		case SNDCTL_DSP_POST:
			dsproxy_audio_emit_ioctl(DSPROXY_STREAM_IOCTL, file, em, cmd, 0);
			return(0);
		
		case SNDCTL_DSP_GETCAPS:
			val = DSP_CAP_REALTIME;  /* no special flags */
			break;
		
		case SNDCTL_DSP_SYNC:
			if (file->f_flags & O_WRONLY)  /* only sync on the write half of the pipe! */
			for (;;) {
				int len;
				dsproxy_audio_flush(em, file);
				len = em->pcm_outstanding;
				if (!len) break;
				interruptible_sleep_on(&em->outq);
				if(signal_pending(current)) return(-ERESTARTSYS);
			}
			return(0);
		
		case SNDCTL_DSP_GETOPTR:
		{
			int rc = 0;
			
			down(&em->sem);
			if ((em->firstwrite.tv_sec == 0) && (em->firstwrite.tv_usec == 0)) {
				c_info.bytes = 0;
			} else {
				do_gettimeofday(&now);
				c_info.bytes = (((now.tv_usec - em->firstwrite.tv_usec) / 10000) +
					100 * (now.tv_sec - em->firstwrite.tv_sec)) * (em->bytes_per_sample * em->channels) * 
					((em->freq / 100) > 0 ? (em->freq / 100) : 1);
				DEBUG(printk("bytes: %lu.%lu: %d\n", now.tv_sec, now.tv_usec, c_info.bytes));
			}
			c_info.blocks = 0;
			c_info.ptr = 0;
			if (copy_to_user((void *)arg, &c_info, sizeof(c_info))) rc = -EFAULT;
			up(&em->sem);
			
			return(rc);
		}
		
		case SNDCTL_DSP_GETOSPACE:
		{
			struct audio_buf_info info;
			int fsiz;
		
			down(&em->sem);
			fsiz = (em->fragsize * em->numfrags);
			info.bytes = get_fifo_avail(em) - DSPROXY_HDRSIZ;
			if (info.bytes > fsiz) info.bytes = fsiz;
			
			info.fragstotal = em->numfrags;
			info.fragments = (em->fragsize > 0) ? info.bytes / em->fragsize : 0;
			info.fragsize = em->fragsize;
			info.bytes = info.fragments * info.fragsize;	
			up(&em->sem);
			
			if (info.fragments > info.fragstotal) info.fragments = info.fragstotal;
			if (copy_to_user((void *)arg, &info, sizeof(info))) {
				DEBUG(printk("can't copy_to_user()\n"));
				return(-EFAULT);
			}
			return(0);
		}
		
		case SNDCTL_DSP_SETFRAGMENT: 
			if (get_user(val, (int *)arg)) return(-EFAULT);
			
			down(&em->ioctl_sem);  /* doesn't need full protection..just to keep fragsize+numfrags atomic */
			em->frag_explicitly_set = 1;
			em->fragsize = 1 << (val & 0x0000ffff);
			if (em->fragsize < 256) em->fragsize = 256;
			else if (em->fragsize > DSPROXY_FRAGFIFO_SIZE) em->fragsize = DSPROXY_FRAGFIFO_SIZE;
			
			em->numfrags = (val >> 16) & 0x0000FFFF;
			if (em->numfrags * em->fragsize > DSPROXY_FRAGFIFO_SIZE) em->numfrags = DSPROXY_FRAGFIFO_SIZE / em->fragsize;
			if (em->numfrags < 1) {
				em->numfrags = 1;
				if (em->numfrags * em->fragsize > DSPROXY_FRAGFIFO_SIZE) em->fragsize = DSPROXY_FRAGFIFO_SIZE;
			}
			em->bufsiz = em->numfrags * em->fragsize;
			val = dsproxy_calc_oss_optfrag(em);
			up(&em->ioctl_sem);
			
			dsproxy_audio_emit_ioctl(DSPROXY_STREAM_IOCTL, file, em, cmd, val);
			
			DEBUG(printk("dsproxy_audio.o: fragsize: %d, nfrags: %d, uid %d\n", em->fragsize, em->numfrags, em->uid));
			break;
		
		case SNDCTL_DSP_SPEED:
			if (get_user(val, (int *)arg)) return(-EFAULT);
			if (val < 1) {
				val = 1;
			} else if (val > 44100) {
				val = 44100;
			}
			dsproxy_audio_flush(em, file);
			
			down(&em->ioctl_sem);
			em->freq = val;
			dsproxy_calc_low_water(em);
			dsproxy_calc_oss_optfrag(em);
			up(&em->ioctl_sem);
			
			dsproxy_audio_emit_ioctl(DSPROXY_STREAM_IOCTL, file, em, cmd, val);
			break;
		
		case SNDCTL_DSP_GETBLKSIZE:
			down(&em->ioctl_sem);
			val = em->fragsize;
			up(&em->ioctl_sem);
			break;
		
		case SNDCTL_DSP_STEREO:
			if (get_user(val, (int *)arg)) return(-EFAULT);
			if ((val > 1) || (val < 0)) return(-EINVAL);
			dsproxy_audio_flush(em, file);
			down(&em->ioctl_sem);
			em->channels = val + 1;
			dsproxy_calc_oss_optfrag(em);
			up(&em->ioctl_sem);
			dsproxy_audio_emit_ioctl(DSPROXY_STREAM_IOCTL, file, em, cmd, val);
			break;
		
		case SNDCTL_DSP_CHANNELS:
			if (get_user(val, (int *)arg)) return(-EFAULT);
			if ((val > 2) || (val < 1)) return(-EINVAL);
			dsproxy_audio_flush(em, file);
			down(&em->ioctl_sem);
			em->channels = val;
			dsproxy_calc_oss_optfrag(em);
			up(&em->ioctl_sem);
			dsproxy_audio_emit_ioctl(DSPROXY_STREAM_IOCTL, file, em, cmd, val);
			break;
		
		case SNDCTL_DSP_SETFMT:
			if (get_user(val, (int *)arg)) return(-EFAULT);
			if (val == AFMT_QUERY) {
				val = em->format;
			} else if (val & DSPROXY_CAPABILITIES) {
				dsproxy_audio_flush(em, file);
				dsproxy_audio_emit_ioctl(DSPROXY_STREAM_IOCTL, file, em, cmd, val);
				down(&em->ioctl_sem);  /* doesn't need full protection..this is just to keep format+bytes atomic */
				em->format = val;
				if (val & (AFMT_MU_LAW | AFMT_U8 | AFMT_S8)) {
					em->bytes_per_sample = 1;
				} else {
					em->bytes_per_sample = 2;
				}
				dsproxy_calc_oss_optfrag(em);
				up(&em->ioctl_sem);
			} else {
				return(-EINVAL);
			}
			break;
		
		case SNDCTL_DSP_GETFMTS:
			if (get_user(val, (int *)arg)) return(-EFAULT);
			val = DSPROXY_CAPABILITIES;
			break;
		
		default:
			DEBUG(printk("dsproxy_audio_ioctl(%08X) unknown\n", cmd));
			return(-EINVAL);
	}
	
	return(put_user(val, (int *)arg));
}


/*
 * dsp read()
 */ 
static int dsproxy_audio_read_2(
	struct file *file, 
	struct dsproxy_s *em, 
	char *buf, 
	size_t count, 
	loff_t *ppos)
{
	int wrapsiz;
	
	DEBUG(printk("dsproxy_audio_read_2() in\n"));
	if (count > 0) {
		DEBUG(printk("dsproxy_audio_read_2() count > 0\n"));
		down(&em->sem);
		
		if (count > em->fifocount) count = em->fifocount;
		
		if (count > 0) {
			if (em->fifoget + count >= DSPROXY_FIFO_SIZE) {
				wrapsiz = DSPROXY_FIFO_SIZE - em->fifoget;		
				if (copy_to_user(buf, em->vmbuf + em->fifoget, wrapsiz)) {
					count = -EFAULT;
				} else {
					em->fifocount -= wrapsiz;
					em->fifoget = 0;
					count=wrapsiz;
				}
			} else {
				if (copy_to_user(buf, em->vmbuf + em->fifoget, count)) {
					count = -EFAULT;
				} else {
					em->fifoget += count;
					em->fifocount -= count;
				}
			}
			
			if (count != -EFAULT) {
				em->pcm_outstanding -= count;
				em->bytes_processed += (unsigned long long)count;
			}
		}
	
		up(&em->sem);
	}

	DEBUG(printk("dsproxy_audio_read_2() out count: %d\n", count));
	return(count);
}


int dsproxy_audio_read(
	struct file *file, 
	struct dsproxy_s *em, 
	char *buf, 
	size_t count, 
	loff_t *ppos)
{
	int dec = 0;
	int err;
	
	DEBUG(printk("dsproxy_audio_read() in\n"));
	
	if (count < 1 || count > 65536 || buf == 0)
		return(-EFAULT);
		
	DEBUG(printk("dsproxy_audio_read() count ok\n"));
	
	if (((err = verify_area(VERIFY_WRITE, (void *)buf, count)) < 0))
		return(err);
	
	DEBUG(printk("dsproxy_audio_read() verified ok\n"));
	
	while (count) {
		dec = dsproxy_audio_read_2(file, em, buf, count, ppos);
		if (dec) {
			if (dec!=-EFAULT) {
				DEBUG(printk("dsproxy_audio_read() dec != -EFAULT\n"));
				wake_up_interruptible(&em->outq);
			}
			break;
		} else {
			DEBUG(printk("dsproxy_audio_read() dec == -EFAULT\n"));
			if (file->f_flags&O_NONBLOCK) {
				DEBUG(printk("dsproxy_audio_read() out EAGAIN\n"));
				return(-EAGAIN);
			}
			DEBUG(printk("dsproxy_audio_read() pre intsleepon\n"));
			interruptible_sleep_on(&em->inq);
			DEBUG(printk("dsproxy_audio_read() post intsleepon\n"));
			if(signal_pending(current)) {
				DEBUG(printk("dsproxy_audio_read() out ERESTARTSYS\n"));
				return(-ERESTARTSYS);
			}
		}
	}
	
	DEBUG(printk("dsproxy_audio_read() out ok\n"));
	
	return(dec);
}


/*
 * dsp write()
 */ 
static int dsproxy_audio_write_2(
	struct dsproxy_s *em, 
	const char *buf, 
	size_t count, 
	int lowwater)
{
	int fsiz;
	int hdrsiz;
	int wrapsiz;
	
	if (count > 0) {
		down(&em->sem);
		
		hdrsiz = DSPROXY_HDRSIZ;  /* header + len compared to frag to keep from */
		if ( (fsiz = get_fifo_avail(em) - hdrsiz) > lowwater ) {  /* thrashing with reads */
			if (count > fsiz) count = fsiz;
			
			fifo_emit_char(em, DSPROXY_STREAM_DSPDATA);  /* emit this fragment's header */
			fifo_emit_short(em, count);
			
			em->hdrskip += (unsigned long long)hdrsiz;
			
			if (em->fifoput + count > DSPROXY_FIFO_SIZE) {
				wrapsiz = DSPROXY_FIFO_SIZE - em->fifoput;		
				memcpy(em->vmbuf + em->fifoput, buf, wrapsiz);
				em->fifocount += count;
				fsiz = count-wrapsiz;
				memcpy(em->vmbuf, buf+wrapsiz, fsiz);
				em->fifoput = fsiz;
			} else {
				memcpy(em->vmbuf + em->fifoput, buf, count);
				em->fifocount+=count;
				em->fifoput+=count;
				if (em->fifoput == DSPROXY_FIFO_SIZE) em->fifoput = 0;
			}
			
			em->pcm_outstanding+=count;
		} else {
			count = 0;  /* not enough fifo room to queue up our request */
		}
		up(&em->sem);
	}
	
	return(count);
}


int dsproxy_audio_write(
	struct file *file, 
	struct dsproxy_s *em, 
	const char *buf, 
	size_t count, 
	loff_t *ppos)
{
	int rc, count2, offset_orig, err;
	int volume, bass, treble, pcm, ogain, dirty;
	
	if (count < 1 || count > 65536 || buf == 0) return(-EFAULT);
	if (((err = verify_area(VERIFY_READ, (void *)buf, count)) < 0)) return(err);
	
	if (count < 0) count = 0; 
	else if (count > DSPROXY_FAULTBUF_SIZ) count = DSPROXY_FAULTBUF_SIZ;
	count2 = count;
	
	down(&em->wr_sem);
	if (copy_from_user(em->faultbuf + (offset_orig = em->offset), buf, count)) {
		up(&em->wr_sem);
		return(-EFAULT);
	} else {
		count2 = em->offset + count;
		if (count2 < em->lowwater) {
			em->offset = count2;
			up(&em->wr_sem);
			return(count);
		} else {
			em->offset = 0;
		}
	}
	
	/*
	 * note that em->sem can't be downed yet since it's downed in dsproxy_audio_emit_ioctl2!
	 * we leave that for dsproxy_audio_write_bytes().  to lock the mixer, we control it via 
	 * a semaphore and take a snapshot of it.  if we get an "old" value, we should be close
	 * enough.
	 */
	down(&em->mixer_sem);
	if (em->dirty&DSPROXY_SOUND_MASK) {
		dirty = em->dirty;
		volume = em->volume;
		bass = em->bass;
		treble = em->treble;
		pcm = em->pcm;
		ogain = em->ogain;
		em->dirty &= (~DSPROXY_SOUND_MASK);
		up(&em->mixer_sem);
		
		if (dirty&SOUND_MASK_VOLUME) dsproxy_audio_emit_ioctl(DSPROXY_MIXER_IOCTL,file,em,SOUND_MIXER_WRITE_VOLUME,volume);
		if (dirty&SOUND_MASK_BASS)   dsproxy_audio_emit_ioctl(DSPROXY_MIXER_IOCTL,file,em,SOUND_MIXER_WRITE_BASS,  bass  );
		if (dirty&SOUND_MASK_TREBLE) dsproxy_audio_emit_ioctl(DSPROXY_MIXER_IOCTL,file,em,SOUND_MIXER_WRITE_TREBLE,treble);
		if (dirty&SOUND_MASK_PCM)    dsproxy_audio_emit_ioctl(DSPROXY_MIXER_IOCTL,file,em,SOUND_MIXER_WRITE_PCM,   pcm   );
		if (dirty&SOUND_MASK_OGAIN)  dsproxy_audio_emit_ioctl(DSPROXY_MIXER_IOCTL,file,em,SOUND_MIXER_WRITE_OGAIN, ogain );
	} else {
		up(&em->mixer_sem);
	}
	
	rc = dsproxy_audio_write_bytes(em, file, count2, offset_orig, em->lowwater, 0);
	up(&em->wr_sem);
	
	return((rc < 0) ? rc : count);
}


int dsproxy_audio_write_bytes(
	struct dsproxy_s *em, 
	struct file *file, 
	int count2, 
	int offset_orig, 
	int lowwater, 
	char ignore_nonblock)
{
	int dec;
	char *buf;
	int rc = 0;
	
	buf = em->faultbuf;  /* this keeps spurious page faults from happening since we own all the data in kernel space */
	while ((count2) && (em->r_inuse)) {
		dec = dsproxy_audio_write_2(em, buf, count2, lowwater);
		if (dec) {
			buf += dec;
			count2 -= dec;
			wake_up_interruptible(&em->inq);
			
			down(&em->ioctl_sem);
			if ((em->firstwrite.tv_sec == 0) && (em->firstwrite.tv_usec == 0)) {
				do_gettimeofday(&em->firstwrite);
			}
			up(&em->ioctl_sem);
		} else {
			if ((!ignore_nonblock) && (file->f_flags & O_NONBLOCK)) {
				em->offset = offset_orig;  /* restore old coalesce offset */
				return(-EAGAIN);
			}
			if (em->r_inuse) {
				interruptible_sleep_on(&em->outq);
				/*
				 * originally commented out since this causes sync problems 
				 * when ctrl-c/z is pressed! but now we defer reporting until 
				 * write is finished.  (i.e., do the return later)
				 */
				if (signal_pending(current)) rc = (-ERESTARTSYS);
			} else {
				break;	/* bail since the write's going to get scrapped anyway */
			}
		}
	}
	
	return(rc);
}


void dsproxy_audio_flush(
	struct dsproxy_s *em, 
	struct file *file)
{
	int len;
	
	down(&em->wr_sem);
	len = em->offset;
	em->offset = 0;
	if (len) dsproxy_audio_write_bytes(em, file, len, 0, 0, 1); /* flush out the stream */
	up(&em->wr_sem);
}


int dsproxy_audio_open(
	struct inode* inode, 
	struct file* filp) 
{
	struct dsproxy_s *em;
	int subdevice;
	
	subdevice = MINOR(inode->i_rdev);
	
	if (((filp->f_flags&O_ACCMODE) != O_WRONLY)&&((filp->f_flags&O_ACCMODE) != O_RDONLY)) {
		DEBUG(printk("dsproxy_main.o: open() access mode %08x is unsupported\n", filp->f_flags));
		return(-EINVAL);
	}
	
	down(&global_sem);
	em = emhead;
	while(em) {
		if (em->uid == current->uid) break;
		em = em->next;
	}
	
	if (!em) {
		if ((filp->f_flags&O_ACCMODE) == O_WRONLY) {
			up(&global_sem);
			return(-EBUSY);  /* disallow open for writes if reader is missing */
		} else {
			em = kmalloc(sizeof(struct dsproxy_s), GFP_KERNEL);  
				/* this may cause slowdown in other processes by causing
				   blocking until the kernel can get ram..hard to tell */
			
			if (!em) {
				up(&global_sem);
				return(-ENOMEM);
			}
			
			memset(em, 0, sizeof(struct dsproxy_s));
			em->next = emhead;
			emhead = em;
			em->uid = current->uid;
			em->subdevice = subdevice;
			sema_init(&em->sem, 1);        /* +1 off zero gives us a mutex (generic r/w lock) */
			sema_init(&em->wr_sem, 1);     /* +1 off zero gives us a mutex (lock for offset coalesce) */
			sema_init(&em->ioctl_sem, 1);  /* +1 off zero gives us a mutex (lock for ioctl touched values) */
			sema_init(&em->mixer_sem, 1);  /* +1 off zero gives us a mutex (lock for mixer volume/dirty values) */
			sema_init(&em->forkread, 1);   /* lock for post fork reads on same fd  */
			sema_init(&em->forkwrite, 1);  /* lock for post fork writes on same fd */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,0)
			init_waitqueue_head(&em->inq);
			init_waitqueue_head(&em->outq);
#endif 
		}
	} else if ((filp->f_flags&O_ACCMODE) == O_RDONLY) {
		up(&global_sem);
		return(-EBUSY);
	}
	
	if (((em->w_inuse) && ((filp->f_flags&O_ACCMODE) == O_WRONLY)) ||  /* disallow open if selected end of pipe is already open */
		((em->r_inuse) && ((filp->f_flags&O_ACCMODE) == O_RDONLY))) {  /* checking reads is overkill, but it doesn't hurt       */
		up(&global_sem);
		return(-EBUSY);
	}
	
	filp->private_data = em;
	
	if ((filp->f_flags&O_ACCMODE) == O_RDONLY) {
		em->r_inuse = 1;
		
		down(&em->mixer_sem);
        em->volume = DSPROXY_MIXER_STEREO_VOLUME(75);
        em->bass = DSPROXY_MIXER_STEREO_VOLUME(75);  
        em->treble = DSPROXY_MIXER_STEREO_VOLUME(75);
        em->pcm = DSPROXY_MIXER_STEREO_VOLUME(75);   
        em->ogain = DSPROXY_MIXER_STEREO_VOLUME(75); 
        em->dirty = SOUND_MASK_VOLUME | SOUND_MASK_BASS | SOUND_MASK_TREBLE | SOUND_MASK_PCM | SOUND_MASK_OGAIN;
        up(&em->mixer_sem);
        
		em->inode=inode;
	} else if ((filp->f_flags&O_ACCMODE) == O_WRONLY) {
		down(&em->ioctl_sem);
		em->w_inuse = 1;
		em->numfrags = 2;
		em->fragsize = DSPROXY_FRAGFIFO_SIZE  / em->numfrags;
		em->frag_explicitly_set = 0;
		em->bufsiz = em->numfrags * em->fragsize;
		em->format = AFMT_U8;
		em->offset = 0;
		em->freq=8000;
		dsproxy_calc_low_water(em);
		em->pcm_outstanding = 0;
		em->bytes_processed = 0ULL;
		em->channels = 1;
		em->bytes_per_sample = 1;
		em->hdrskip = 0ULL;
		em->firstwrite.tv_sec = 0;
		em->firstwrite.tv_usec = 0;
		em->send_reset = 1;
		up(&em->ioctl_sem);
		
		/*
		 * note that fifoget/put/etc are conspicuously absent from
		 * being initialized.  this is so we don't pull the rug out
		 * from under the reader process.  they're implicitly set
		 * to zero the first time through when we kmalloc then
		 * memset our dsproxy_s struct when the reader is opened!
		 * since we disallow write opens if the reader isn't opened,
		 * this is OK.
		 */
	}
	
	/*
	 * this is tricky and normally not recommended.  this ups the inode sem counts such that
	 * we can virtualize the device.  note that this implies that we do *all* our own
	 * bookkeeping locally
	 */
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	up(&inode->i_atomic_write);  /* increment the atomic write count */
	#else 
	inode->i_count.counter++;  /* increment the atomic write count */
	#endif
	up(&inode->i_sem);  /* increment the sem count of the inode */
	
	em->clients++;
	up(&global_sem);
	
	DEBUG(printk("dsproxy_audio.o: Opening device for uid %d (dsproxy_s -> %p)\n", current->uid,em));
	
	return(0);
}

int dsproxy_audio_release(
	struct dsproxy_s *em, 
	struct file *filp, 
	int subdevice)
{
	struct dsproxy_s *t;
	
	if ((filp->f_flags&O_ACCMODE) == O_WRONLY) {
		
		/* THIS CAN CAUSE DEADLOCK CONDITIONS
		 *	dsproxy_audio_flush(em, filp);
		 *      dsproxy_audio_emit_ioctl(DSPROXY_STREAM_IOCTL, filp, em, SNDCTL_DSP_RESET, 0);
		 * THIS CAN CAUSE DEADLOCK CONDITIONS */
		
		down(&global_sem);  /* this is as far up as we want/need to put this leg of the top half of the sem pair */
		em->w_inuse = 0;
		DEBUG(printk("dsproxy_main.o: Releasing device %d (write), uid %d\n", subdevice, em->uid));
	} else if ((filp->f_flags&O_ACCMODE) == O_RDONLY) {
		down(&global_sem);				
		em->r_inuse = 0;
		if (em->w_inuse) {
			wake_up_interruptible(&em->outq);  /* let writer process finish last chunk of data if it has to */
		}
		
        DEBUG(printk("dsproxy_main.o: Releasing device %d (read), uid %d\n", subdevice, em->uid));
	} else {
		/* should never happen since we don't allow any other modes for open! */
		down(&global_sem);
		DEBUG(printk("dsproxy_main.o: Releasing device %d (access mode %08x [unknown]), uid %d\n", subdevice, filp->f_flags, em->uid));
	}
	
	em->clients--;
	
	/*
	 * this is tricky and normally not recommended.  this downs the inode sem counts such that
	 * we can de-virtualize the device.  note that this implies that we do *all* our own
	 * bookkeeping locally
	 */
	#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,0)
	down(&em->inode->i_atomic_write);  /* decrement the inflated atomic write count */
	#else 
	em->inode->i_count.counter--;  /* decrement the inflated atomic write count */
	#endif
	down(&em->inode->i_sem);  /* decrement the inflated sem count of the inode*/
	
	if (!em->clients) {
		t = emhead;  /* prune out this uid's device instance */
		if (t == em) {
			emhead = t->next;
		} else {
			while(t) {
				if (t->next == em) {
					t->next = em->next;
					break;
				}
				t = t->next;
			}
		}
		
		DEBUG(printk("dsproxy_main.o: freeing device for uid %d (dsproxy_s -> %p)\n", em->uid, em));
		kfree(em);
	}
	
	up(&global_sem);  /* bottom half of critical section */
	
	return(0);
}


unsigned int dsproxy_audio_poll(
	struct file *filp, 
	poll_table *wait)
{
	struct dsproxy_s *em = filp->private_data;
	unsigned int mask = 0;
	
	poll_wait(filp, &em->inq, wait);
	poll_wait(filp, &em->outq, wait);
	
	down(&em->sem);
	if (em->fifocount > 0) mask |= POLLIN | POLLRDNORM;  /* read data is available */
	if (get_fifo_avail(em) > DSPROXY_HDRSIZ) mask |= POLLOUT | POLLWRNORM;  /* write buffer space is available */
	up(&em->sem);
	
	return(mask);
}
